#include <chrono>
#include <cstdlib>
#include <iostream>

/* MPI header */
#include <mpi.h>

#include "is-prime.h"

int main(int argc, char** argv)
{
   MPI_Init(&argc, &argv);
   
   // how many processes are there?
   int size = 0;  
   MPI_Comm_size(MPI_COMM_WORLD, &size); 
   
   // what is the rank of this process?
   int rank = 0;
   MPI_Comm_rank(MPI_COMM_WORLD, &rank); 
   
   // read up to what max number we are going to count the prime numbers
   int64_t limit = std::atoi(argv[1]);
   
   if(!rank)
   {
      std::cout << "How many prime numbers p are there with p < "
                << limit << "? (" << size << " processes)\n\n";
      std::cout.flush();
   }
   
   int64_t counted_primes = 0;
   
   // check 6k-1, starting from 5
   for(int64_t n = 6*(rank+1) - 1; n < limit; n += 6*size)
      if(is_prime(n)) counted_primes++;

   // check 6k+1, starting from 7
   for(int64_t n = 6*(rank+1) + 1; n < limit; n += 6*size)
      if(is_prime(n)) counted_primes++;
      
   int64_t overall_primes = 0;
   
   std::clog << "\tRank " << rank << " counted\t" << counted_primes << "\tprimes.\n";
   std::clog.flush();
   
   /*
    * int MPI_Reduce(
    *    const void *sendbuf, void *recvbuf, int count,
    *    MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm
    * );
    * See documentation: https://www.open-mpi.org/doc/v4.1/man3/MPI_Reduce.3.php
    */
   MPI_Reduce(&counted_primes, &overall_primes, 1, MPI_INT64_T, MPI_SUM, 0, MPI_COMM_WORLD);
   if(!rank)
   {
      std::clog << "\tThere are also\t2\tprimes below five.\n";
      overall_primes += 2;
      std::clog.flush();
   }

   MPI_Finalize();
   
   if(!rank) std::cout << "\nTotal: There are " << overall_primes << " prime numbers smaller than " << limit << ".\n";
}
